@@ -10,6 +10,7 @@ import (
 	"os"
 	"sort"
 	"strings"
+	"time"
 
 	"github.com/ghodss/yaml"
 	"github.com/sirupsen/logrus"
@@ -22,6 +23,9 @@ import (
 	"github.com/openshift/installer/pkg/types"
 )
 
+const localhostKubeApiServerUrl string = "https://localhost:6443"
+const bootstrapIpServerPort string = "64444"
+
 var log = logrus.New()
 
 type Cluster struct {
@@ -55,6 +59,10 @@ type ApiLBConfig struct {
 	FrontendAddr string
 }
 
+type IngressConfig struct {
+	Peers []string
+}
+
 type Node struct {
 	Cluster           Cluster
 	LBConfig          ApiLBConfig
@@ -63,6 +71,9 @@ type Node struct {
 	EtcdShortHostname string
 	VRRPInterface     string
 	DNSUpstreams      []string
+	BootstrapIP       string
+	IngressConfig     IngressConfig
+	EnableUnicast     bool
 }
 
 func getDNSUpstreams(resolvConfPath string) (upstreams []string, err error) {
@@ -170,17 +181,100 @@ func (c *Cluster) PopulateVRIDs() error {
 	}
 	return nil
 }
-
-func GetConfig(kubeconfigPath, clusterConfigPath, resolvConfPath string, apiVip net.IP, ingressVip net.IP, dnsVip net.IP, apiPort, lbPort, statPort uint16) (node Node, err error) {
-	// Try cluster-config.yml first
-	clusterName, clusterDomain, err := getClusterConfigClusterNameAndDomain(clusterConfigPath)
+func GetBootstrapIP(apiVip string) (bootstrapIP string, err error) {
+	conn, err := net.DialTimeout("tcp", net.JoinHostPort(apiVip, bootstrapIpServerPort), 10*time.Second)
 	if err != nil {
-		// We are using kubeconfig as a fallback for this
-		clusterName, clusterDomain, err = GetKubeconfigClusterNameAndDomain(kubeconfigPath)
-		if err != nil {
-			return node, err
+		log.Debugf("An error occurred on dial: %v", err)
+		return "", err
+	}
+	defer conn.Close()
+
+	bootstrapIP, err = bufio.NewReader(conn).ReadString('\n')
+	if err != nil {
+		log.Infof("An error occurred on read: %v", err)
+		return "", err
+	}
+
+	bootstrapIP = strings.TrimSpace(bootstrapIP)
+
+	log.Infof("Got bootstrap IP %v", bootstrapIP)
+
+	return bootstrapIP, err
+}
+
+func GetVRRPConfig(apiVip, ingressVip, dnsVip net.IP) (vipIface net.Interface, nonVipAddr *net.IPNet, err error) {
+	vips := make([]net.IP, 0)
+	if apiVip != nil {
+		vips = append(vips, apiVip)
+	}
+	if ingressVip != nil {
+		vips = append(vips, ingressVip)
+	}
+	if dnsVip != nil {
+		vips = append(vips, dnsVip)
+	}
+	return GetInterfaceAndNonVIPAddr(vips)
+}
+
+func IsUpgradeStillRunning(kubeconfigPath string) (error, bool) {
+
+	config, err := clientcmd.BuildConfigFromFlags("", kubeconfigPath)
+	if err != nil {
+		return err, true
+	}
+
+	clientset, err := kubernetes.NewForConfig(config)
+	if err != nil {
+		return err, true
+	}
+
+	nodes, err := clientset.CoreV1().Nodes().List(metav1.ListOptions{})
+	if err != nil {
+		return err, true
+	}
+
+	for _, node := range nodes.Items {
+		if node.Annotations["machineconfiguration.openshift.io/desiredConfig"] != node.Annotations["machineconfiguration.openshift.io/currentConfig"] {
+			return nil, true
 		}
 	}
+
+	return nil, false
+}
+
+func GetIngressConfig(kubeconfigPath string) (ingressConfig IngressConfig, err error) {
+	config, err := clientcmd.BuildConfigFromFlags("", kubeconfigPath)
+	if err != nil {
+		return ingressConfig, err
+	}
+
+	clientset, err := kubernetes.NewForConfig(config)
+	if err != nil {
+		return ingressConfig, err
+	}
+
+	nodes, err := clientset.CoreV1().Nodes().List(metav1.ListOptions{})
+	if err != nil {
+		return ingressConfig, err
+	}
+
+	for _, node := range nodes.Items {
+		for _, address := range node.Status.Addresses {
+			if address.Type == v1.NodeInternalIP {
+				ingressConfig.Peers = append(ingressConfig.Peers, address.Address)
+			}
+		}
+	}
+
+	return ingressConfig, nil
+}
+
+func GetConfig(kubeconfigPath, clusterConfigPath, resolvConfPath string, apiVip net.IP, ingressVip net.IP, dnsVip net.IP, apiPort, lbPort, statPort uint16) (node Node, err error) {
+	clusterName, clusterDomain, err := GetClusterNameAndDomain(kubeconfigPath, clusterConfigPath)
+	if err != nil {
+		return node, err
+	}
+
 	node.Cluster.Name = clusterName
 	node.Cluster.Domain = clusterDomain
 
@@ -205,12 +299,9 @@ func GetConfig(kubeconfigPath, clusterConfigPath, resolvConfPath string, apiVip
 	if err != nil {
 		return node, err
 	}
-
-	vips := make([]net.IP, 0)
 	node.Cluster.APIVIPRecordType = "A"
 	node.Cluster.APIVIPEmptyType = "AAAA"
 	if apiVip != nil {
-		vips = append(vips, apiVip)
 		node.Cluster.APIVIP = apiVip.String()
 		if apiVip.To4() == nil {
 			node.Cluster.APIVIPRecordType = "AAAA"
@@ -220,7 +311,6 @@ func GetConfig(kubeconfigPath, clusterConfigPath, resolvConfPath string, apiVip
 	node.Cluster.IngressVIPRecordType = "A"
 	node.Cluster.IngressVIPEmptyType = "AAAA"
 	if ingressVip != nil {
-		vips = append(vips, ingressVip)
 		node.Cluster.IngressVIP = ingressVip.String()
 		if ingressVip.To4() == nil {
 			node.Cluster.IngressVIPRecordType = "AAAA"
@@ -228,15 +318,19 @@ func GetConfig(kubeconfigPath, clusterConfigPath, resolvConfPath string, apiVip
 		}
 	}
 	if dnsVip != nil {
-		vips = append(vips, dnsVip)
 		node.Cluster.DNSVIP = dnsVip.String()
 	}
-	vipIface, nonVipAddr, err := getInterfaceAndNonVIPAddr(vips)
+	vipIface, nonVipAddr, err := GetVRRPConfig(apiVip, ingressVip, dnsVip)
 	if err != nil {
 		return node, err
 	}
 	node.NonVirtualIP = nonVipAddr.IP.String()
 
+	node.EnableUnicast = false
+	if os.Getenv("ENABLE_UNICAST") == "yes" {
+		node.EnableUnicast = true
+	}
+
 	resolvConfUpstreams, err := getDNSUpstreams(resolvConfPath)
 	if err != nil {
 		return node, err
@@ -254,8 +348,11 @@ func GetConfig(kubeconfigPath, clusterConfigPath, resolvConfPath string, apiVip
 		return node, errors.New("No upstream DNS servers found")
 	}
 
-	prefix, _ := nonVipAddr.Mask.Size()
-	node.Cluster.VIPNetmask = prefix
+	if apiVip.To4() == nil {
+		node.Cluster.VIPNetmask = 128
+	} else {
+		node.Cluster.VIPNetmask = 32
+	}
 	node.VRRPInterface = vipIface.Name
 
 	// We can't populate this with GetLBConfig because in many cases the
@@ -269,9 +366,15 @@ func GetConfig(kubeconfigPath, clusterConfigPath, resolvConfPath string, apiVip
 	return node, err
 }
 
-func getSortedBackends(kubeconfigPath string) (backends []Backend, err error) {
+// getSortedBackends builds config to communicate with kube-api based on kubeconfigPath parameter value, if kubeconfigPath is not empty it will build the
+// config based on that content else config will point to localhost.
+func getSortedBackends(kubeconfigPath string, readFromLocalAPI bool) (backends []Backend, err error) {
 
-	config, err := clientcmd.BuildConfigFromFlags("", kubeconfigPath)
+	kubeApiServerUrl := ""
+	if readFromLocalAPI {
+		kubeApiServerUrl = localhostKubeApiServerUrl
+	}
+	config, err := clientcmd.BuildConfigFromFlags(kubeApiServerUrl, kubeconfigPath)
 	if err != nil {
 		log.WithFields(logrus.Fields{
 			"err": err,
@@ -321,19 +424,24 @@ func GetLBConfig(kubeconfigPath string, apiPort, lbPort, statPort uint16, apiVip
 		LbPort:   lbPort,
 		StatPort: statPort,
 	}
+
 	// LB frontend address: IPv6 '::' , IPv4 ''
 	if apiVip.To4() == nil {
 		config.FrontendAddr = "::"
 	}
-
-	backends, err := getSortedBackends(kubeconfigPath)
+	// Try reading master nodes details first from api-vip:kube-apiserver and failover to localhost:kube-apiserver
+	backends, err := getSortedBackends(kubeconfigPath, false)
 	if err != nil {
-		log.WithFields(logrus.Fields{
-			"kubeconfigPath": kubeconfigPath,
-		}).Error("Failed to retrieve API members information")
-		return config, err
+		log.Infof("An error occurred while trying to read master nodes details from api-vip:kube-apiserver: %v", err)
+		log.Infof("Trying to read master nodes details from localhost:kube-apiserver")
+		backends, err = getSortedBackends(kubeconfigPath, true)
+		if err != nil {
+			log.WithFields(logrus.Fields{
+				"kubeconfigPath": kubeconfigPath,
+			}).Error("Failed to retrieve API members information")
+			return config, err
+		}
 	}
-
 	// The backends port is the Etcd one, but we need to loadbalance the API one
 	for i := 0; i < len(backends); i++ {
 		backends[i].Port = apiPort
@@ -344,3 +452,14 @@ func GetLBConfig(kubeconfigPath string, apiPort, lbPort, statPort uint16, apiVip
 	}).Debug("Config for LB configuration retrieved")
 	return config, nil
 }
+
+func GetClusterNameAndDomain(kubeconfigPath, clusterConfigPath string) (clusterName string, clusterDomain string, err error) {
+	// Try cluster-config.yml first
+	clusterName, clusterDomain, err = getClusterConfigClusterNameAndDomain(clusterConfigPath)
+	if err != nil {
+		// We are using kubeconfig as a fallback for this
+		clusterName, clusterDomain, err = GetKubeconfigClusterNameAndDomain(kubeconfigPath)
+	}
+
+	return
+}
